import itertools
import os
import pathlib
import tempfile

import gin
import pytest
import torch

from msprior.scripted import ScriptedPrior

torch.set_grad_enabled(False)

configs = map(str, pathlib.Path("msprior/configs").glob("*.gin"))
configs = list(configs)

names = map(os.path.basename, configs)
names = map(lambda x: os.path.splitext(x)[0], names)

names = itertools.product(names, ["continuous", "discrete"],
                          ["listen", "generate"])
names = map(lambda x: " ".join(x), names)

continuous = [True, False]
listen = [True, False]
configs = itertools.product(configs, continuous, listen)


@pytest.mark.parametrize("config,continuous,listen", configs, ids=names)
def test_config(config, continuous, listen):
    gin.clear_config()
    gin.parse_config_file(config)

    model = ScriptedPrior(
        from_continuous=continuous,
        initial_listen=listen,
    )

    if listen and not continuous:
        x = torch.randint(0, 16, (1, model.forward_params[0], 8)).float()
        y = model(x)
        assert torch.allclose(y[:, :-1].float(), x[:, :y.shape[1] - 1].float())

        x = torch.randint(0, 16, (4, model.forward_params[0], 8)).float()
        y = model(x)
        assert torch.allclose(y[:, :-1].float(), x[:, :y.shape[1] - 1].float())

    with tempfile.TemporaryDirectory() as tmp:
        model.export_to_ts(os.path.join(tmp, "model.ts"))



import importlib
import sys

from absl import app

AVAILABLE_SCRIPTS = ['preprocess', 'train', 'export', 'compact']


def help():
    print(f"""usage: msprior [ {' | '.join(AVAILABLE_SCRIPTS)} ]

positional arguments:
  command     Command to launch with msprior.
""")
    exit()


def main():
    if len(sys.argv) == 1:
        help()
    elif sys.argv[1] not in AVAILABLE_SCRIPTS:
        help()

    command = sys.argv[1]

    module = importlib.import_module("msprior_scripts." + command)
    sys.argv[0] = module.__name__
    app.run(module.main)



import logging
import os

import gin
import pytorch_lightning as pl
import torch
from absl import app, flags
from pytorch_lightning import callbacks
from torch.utils import data

from msprior.attention import Prior
from msprior.dataset import SequenceDataset

FLAGS = flags.FLAGS
flags.DEFINE_multi_string("config",
                          default="msprior/configs/decoder_only.gin",
                          help="config to parse.")
flags.DEFINE_string("db_path",
                    default=None,
                    required=True,
                    help="path to dataset.")
flags.DEFINE_integer("val_size",
                     default=8192,
                     help="size of validation dataset.")
flags.DEFINE_integer("batch_size", default=64, help="batch size.")
flags.DEFINE_string("name", default=None, required=True, help="train name.")
flags.DEFINE_integer("gpu", default=0, help="gpu index.")
flags.DEFINE_integer("workers",
                     default=0,
                     help="num workers during data loading.")
flags.DEFINE_string("pretrained_embedding",
                    default=None,
                    help="use pretrained embeddings from rave.")
flags.DEFINE_multi_string("override",
                          default=[],
                          help="additional gin bindings.")
flags.DEFINE_string("ckpt",
                    default=None,
                    help="checkpoint to resume training from.")


def add_ext(config: str):
    if config[-4:] != ".gin":
        config += ".gin"
    return config


def main(argv):
    logging.info("parsing configuration")
    configs = list(map(add_ext, FLAGS.config))

    gin.parse_config_files_and_bindings(
        configs,
        [f"PRETRAINED_RAVE='{FLAGS.pretrained_embedding}'"] + FLAGS.override,
    )

    logging.info("building model")
    model = Prior()

    logging.info("loading dataset")
    dataset = SequenceDataset(db_path=FLAGS.db_path)
    train, val = data.random_split(
        dataset,
        (len(dataset) - FLAGS.val_size, FLAGS.val_size),
        generator=torch.Generator().manual_seed(42),
    )
    train_loader = data.DataLoader(
        train,
        batch_size=FLAGS.batch_size,
        shuffle=True,
        drop_last=True,
        num_workers=FLAGS.workers,
    )
    val_loader = data.DataLoader(
        val,
        batch_size=FLAGS.batch_size,
        shuffle=False,
        drop_last=False,
        num_workers=FLAGS.workers,
    )

    os.makedirs(os.path.join("runs", FLAGS.name), exist_ok=False)
    with open(os.path.join("runs", FLAGS.name, "config.gin"),
              "w") as config_out:
        config_out.write(gin.config_str())

    logging.info("creating trainer")
    trainer = pl.Trainer(
        logger=pl.loggers.TensorBoardLogger('runs', name=FLAGS.name),
        accelerator='gpu',
        devices=[FLAGS.gpu],
        callbacks=[
            callbacks.LearningRateMonitor(logging_interval='step'),
            callbacks.ModelCheckpoint(monitor="val_cross_entropy",
                                      filename='best'),
            callbacks.ModelCheckpoint(filename='last'),
            callbacks.EarlyStopping(
                "val_cross_entropy",
                patience=10,
            )
        ],
        log_every_n_steps=10,
        val_check_interval=1000,
    )

    torch.backends.cudnn.benchmark = True
    torch.set_float32_matmul_precision('high')

    logging.info("launch training")
    trainer.fit(
        model,
        train_loader,
        val_loader,
        ckpt_path=FLAGS.ckpt,
    )


if __name__ == "__main__":
    app.run(main)



import functools
import multiprocessing
import os
import pathlib
import subprocess
from itertools import repeat
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Tuple

import gin
import numpy as np
import torch
import torch.nn as nn

# GIN UTILS


def get_feed_forward_size(model_dim):
    return 4 * model_dim


@gin.configurable
def build_warmed_exponential_lr_scheduler(
        optim: torch.optim.Optimizer, start_factor: float, peak_iteration: int,
        decay_factor: float) -> torch.optim.lr_scheduler._LRScheduler:
    linear = torch.optim.lr_scheduler.LinearLR(
        optim,
        start_factor=start_factor,
        end_factor=1.,
        total_iters=peak_iteration,
    )
    exp = torch.optim.lr_scheduler.ExponentialLR(
        optim,
        gamma=decay_factor,
    )
    return torch.optim.lr_scheduler.SequentialLR(optim, [linear, exp],
                                                 milestones=[peak_iteration])


def total_num_tokens(num_quantizers: int, num_tokens: int,
                     num_reserved_tokens: int):
    return num_quantizers * (num_tokens + num_reserved_tokens)


def per_quantizer_num_tokens(num_tokens: int, num_reserved_tokens: int):
    return num_tokens + num_reserved_tokens


def get_num_quantizers(rave_path: str,
                       quantizer_truncation: Optional[int] = None) -> int:
    if quantizer_truncation is not None:
        return quantizer_truncation
    rave = torch.jit.load(rave_path).eval()
    return rave.encode_params[2].item()


# MODEL UTILS


@torch.no_grad()
def perplexity(pred: torch.Tensor):
    pred = torch.log_softmax(pred, -1)
    entropy = -(pred * torch.exp(pred))
    entropy.masked_fill_(torch.isnan(entropy), 0)
    perplexity = torch.exp(entropy.sum(-1)).mean()
    return perplexity


def load_run(run: Optional[str] = None) -> Tuple[str, nn.Module]:

    @gin.configurable
    def get_model(model):
        return model

    print(f'loading run {run}')
    if run is None: return None

    config = os.path.join(run, 'config.gin')
    gin.clear_config()
    gin.parse_config_file(config)
    gin.parse_config(r'get_model.model = %FULL_MODEL')
    model = get_model()

    try:
        ckpt = next(iter(pathlib.Path(run).rglob('best*')))
        ckpt = torch.load(ckpt, map_location="cpu")["state_dict"]
        state = model.state_dict()
        state.update(ckpt)
        model.load_state_dict(state)
    except:
        print(
            f'No compatible ckpt found in {run}, model is left randomly initialised...'
        )
        model = get_model()

    model.eval()

    return run, model


@torch.no_grad()
def pretrained_embedder_init(pretrained_path: str, num_reserved_tokens: int):
    model = torch.jit.load(pretrained_path).eval()
    codebook = model.quantizer.embed
    additional_tokens = torch.randn(
        codebook.shape[0],
        num_reserved_tokens,
        codebook.shape[-1],
    )
    codebook = torch.cat([additional_tokens, codebook], 1)
    codebook = codebook.reshape(-1, codebook.shape[-1])

    embedding = nn.Embedding(*codebook.shape)
    embedding.weight.data.copy_(codebook)
    return embedding


# PREPROCESSING UTILS


def float_array_to_int16_bytes(x):
    return np.floor(x * (2**15 - 1)).astype(np.int16).tobytes()


def load_audio_chunk(
        path: str,
        n_signal: int,
        sr: int,
        silenceremove: Optional[int] = None) -> Iterable[np.ndarray]:

    ffmpeg_cmd = [
        'ffmpeg',
        '-hide_banner',
        '-loglevel',
        'panic',
        '-i',
        path,
        '-ac',
        '1',
        '-ar',
        str(sr),
    ]

    if silenceremove is not None:
        ffmpeg_cmd.extend(["-af", f"silenceremove=1:0:-{silenceremove}dB"])

    ffmpeg_cmd.extend(['-f', 's16le', '-'])
    process = subprocess.Popen(
        ffmpeg_cmd,
        stdout=subprocess.PIPE,
    )
    chunk = process.stdout.read(n_signal * 2)
    while len(chunk) == n_signal * 2:
        chunk = np.frombuffer(chunk, np.int16).astype(np.float32) / 2**15
        yield chunk, path
        chunk = process.stdout.read(n_signal * 2)

    process.stdout.close()


def batch(iterator: Iterable, batch_size: int):
    batch = []
    for elm in iterator:
        batch.append(elm)
        if len(batch) == batch_size:
            yield batch
            batch = []
    if len(batch):
        yield batch


def tuple_as_args(fun):

    def wrapper(tuple_arg):
        return fun(*tuple_arg)

    return wrapper


def flatmap(
    pool: multiprocessing.Pool,
    func: Callable,
    iterable: Iterable,
    queue_size: int,
    chunksize=None,
):
    queue = multiprocessing.Manager().Queue(maxsize=queue_size)
    pool.map_async(
        functools.partial(flat_mappper, func),
        zip(iterable, repeat(queue)),
        chunksize,
        lambda _: queue.put(None),
        lambda *e: print(e),
    )

    item = queue.get()
    while item is not None:
        yield item
        item = queue.get()


def flat_mappper(func, arg):
    data, queue = arg
    for item in func(data):
        queue.put(item)


# PREPROCESSING UTILS


def build_target(inputs: Dict[str, torch.Tensor], input_key: str):
    target_tensor = inputs[input_key].roll(-1, 1)
    target_tensor[:, -1] = -1
    inputs['targets'] = target_tensor


def look_ahead(inputs: Dict[str, torch.Tensor], key: str):
    inputs[key] = inputs[key].roll(-1, 1)
    inputs[key][:, -1] = inputs[key][:, -2]


def compose(functions: Sequence[Callable[[Dict[str, torch.Tensor]], None]]):

    def composed_function(inputs: Dict[str, torch.Tensor]):
        for f in functions:
            f(inputs)

    return composed_function


# OTHERS

# def sample_from_logits(logits: torch.Tensor,
#                        temperature: float = 0.) -> torch.Tensor:
#     assert temperature >= 0
#     dist = torch.softmax(logits / (temperature + 1e-15), -1)
#     dist = dist.reshape(-1, dist.shape[-1])
#     samples = torch.multinomial(dist, 1).reshape_as(logits[..., 0])
#     return samples


def sample_from_logits(logits: torch.Tensor,
                       temperature: float = 0.) -> torch.Tensor:
    if temperature != 0:
        logits = torch.log_softmax(logits / temperature, -1)
        logits -= torch.empty_like(logits).exponential_().log()

    samples = torch.argmax(logits, -1)
    return samples



import dataclasses
from typing import Literal, Optional, Sequence, Union

import numpy as np
import pathos.multiprocessing as mp
import torch
from udls.generated import AudioExample

_PREPROCESSOR_LIST = []
_PREPROCESSOR_DEFAULT_GPU = torch.device('cpu')
_SAMPLING_RATE = 44100

DTYPE_TO_PRECISION = {
    np.int16: AudioExample.Precision.INT16,
    np.int32: AudioExample.Precision.INT32,
    np.int64: AudioExample.Precision.INT64,
    np.float16: AudioExample.Precision.FLOAT16,
    np.float32: AudioExample.Precision.FLOAT32,
    np.float64: AudioExample.Precision.FLOAT64,
}

PRECISION_TO_DTYPE = {
    AudioExample.Precision.INT16: np.int16,
    AudioExample.Precision.INT32: np.int32,
    AudioExample.Precision.INT64: np.int64,
    AudioExample.Precision.FLOAT16: np.float16,
    AudioExample.Precision.FLOAT32: np.float32,
    AudioExample.Precision.FLOAT64: np.float64,
}

Scalar = Union[int, float]
Array = np.ndarray


def interpolate(input_array: np.ndarray, size: int, mode: str) -> np.ndarray:
    input_array = torch.from_numpy(input_array).float()
    assert len(input_array.shape) < 3

    if len(input_array.shape) == 2:
        input_array = input_array.permute(0, 1)
    else:
        input_array = input_array[None, :]

    input_array = input_array[None]

    output_array = torch.nn.functional.interpolate(
        input_array,
        size=size,
        mode=mode,
    )

    output_array = output_array[0].permute(1, 0)

    return output_array.numpy()


@dataclasses.dataclass
class Preprocessor:
    name: str
    mode: Literal['static', 'dynamic']
    dtype: np.dtype
    resample: bool = True
    resampling_mode: Literal['nearest', 'linear'] = 'nearest'

    parallel: bool = False
    device: Optional[torch.device] = None
    sampling_rate: Optional[int] = None

    def __post_init__(self):
        self.device = self.device or get_default_gpu()
        self.sampling_rate = self.sampling_rate or get_sampling_rate()

    def process_single_example(self, audio_array: Array,
                               audio_path: str) -> Union[Scalar, Array]:
        raise NotImplementedError

    def __call__(self, audio_arrays: Sequence[Array],
                 audio_paths: Sequence[str]) -> Sequence[Union[Scalar, Array]]:
        if self.parallel:

            def process(args):
                return self.process_single_example(*args)

            with mp.Pool() as pool:
                outputs = list(
                    pool.map(
                        process,
                        zip(audio_arrays, audio_paths),
                    ))
        else:
            outputs = list(
                map(
                    self.process_single_example,
                    audio_arrays,
                    audio_paths,
                ))

        return outputs


def preprocess_batch(
        audio_arrays: Sequence[Array],
        audio_paths: Sequence[str],
        resampled_seq_length: Optional[int] = None) -> Sequence[AudioExample]:

    audio_examples = [AudioExample() for _ in range(len(audio_arrays))]

    for preprocessor in _PREPROCESSOR_LIST:
        preprocessed = preprocessor(audio_arrays, audio_paths)

        if preprocessor.mode == 'static':
            for ae, label in zip(audio_examples, preprocessed):
                if preprocessor.dtype == str:
                    ae.metadata[preprocessor.name] = str(label)
                else:
                    buffer = ae.buffers[preprocessor.name]
                    buffer.data = np.asarray(label).astype(
                        preprocessor.dtype).tobytes()
                    buffer.precision = DTYPE_TO_PRECISION[preprocessor.dtype]
                    buffer.metadata['mode'] = 'static'
                    buffer.shape.extend([1])

        elif preprocessor.mode == 'dynamic':
            for ae, array in zip(audio_examples, preprocessed):
                if preprocessor.resample and resampled_seq_length is not None:
                    array = interpolate(
                        array,
                        resampled_seq_length,
                        preprocessor.resampling_mode,
                    )
                buffer = ae.buffers[preprocessor.name]

                buffer.data = array.astype(preprocessor.dtype).tobytes()
                buffer.shape.extend(array.shape)
                buffer.precision = DTYPE_TO_PRECISION[preprocessor.dtype]
                buffer.metadata['mode'] = 'dynamic'

    return audio_examples


def register_preprocessor(preprocessor: Preprocessor):
    _PREPROCESSOR_LIST.append(preprocessor)


def get_preprocessors() -> Sequence[Preprocessor]:
    return _PREPROCESSOR_LIST


def set_default_gpu(device: torch.device):
    global _PREPROCESSOR_DEFAULT_GPU
    _PREPROCESSOR_DEFAULT_GPU = device


def get_default_gpu():
    return _PREPROCESSOR_DEFAULT_GPU


def set_sampling_rate(sampling_rate: int):
    global _SAMPLING_RATE
    _SAMPLING_RATE = sampling_rate


def get_sampling_rate():
    return _SAMPLING_RATE
